#include <shogun/base/init.h>
#include <shogun/base/some.h>
#include <shogun/statistical_testing/TestEnums.h>
#include <shogun/statistical_testing/MultiKernelQuadraticTimeMMD.h>
#include <shogun/features/DenseFeatures.h>
#include <shogun/kernel/GaussianKernel.h>
#include <shogun/lib/WrappedObjectArray.h>
#include <shogun/io/CSVFile.h>
#include <shogun/io/SerializableAsciiFile.h>
#include <shogun/kernel/CombinedKernel.h>
#include <shogun/lib/SGVector.h>
#include <shogun/statistical_testing/QuadraticTimeMMD.h>

using namespace shogun;

int main(int, char*[])
{
init_shogun_with_defaults();

auto f_features_p = some<CCSVFile>("../../data/two_sample_test_gaussian.dat");
auto f_features_q = some<CCSVFile>("../../data/two_sample_test_laplace.dat");

//![create_features]
auto features_p = some<CDenseFeatures<float64_t>>(f_features_p);
auto features_q = some<CDenseFeatures<float64_t>>(f_features_q);
//![create_features]

//![create_instance]
auto mmd = some<CQuadraticTimeMMD>(features_p, features_q);
auto kernel = some<CGaussianKernel>(10, 1);
mmd->set_kernel(kernel);
auto alpha = 0.05;
//![create_instance]

//![estimate_mmd]
mmd->set_statistic_type(EStatisticType::ST_BIASED_FULL);
auto statistic = mmd->compute_statistic();
//![estimate_mmd]

//![perform_test]
mmd->set_null_approximation_method(ENullApproximationMethod::NAM_PERMUTATION);
mmd->set_num_null_samples(200);
auto threshold = mmd->compute_threshold(alpha);
auto p_value = mmd->compute_p_value(statistic);
//![perform_test]

//![add_kernels]
auto kernel1 = some<CGaussianKernel>(10, 0.1);
auto kernel2 = some<CGaussianKernel>(10, 1);
auto kernel3 = some<CGaussianKernel>(10, 10);
mmd->add_kernel(kernel1);
mmd->add_kernel(kernel2);
mmd->add_kernel(kernel3);
//![add_kernels]

//![multi_kernel]
auto mk = mmd->multikernel();
mk->add_kernel(kernel1);
mk->add_kernel(kernel2);
mk->add_kernel(kernel2);

auto mk_statistic = mk->compute_statistic();
auto mk_p_value = mk->compute_p_value();
//![multi_kernel]

//![enable_train_test_mode]
mmd->set_train_test_mode(true);
mmd->set_train_test_ratio(1);
//![enable_train_test_mode]

//![select_kernel_single]
auto num_runs = 1;
auto num_folds = 3;
mmd->set_kernel_selection_strategy(EKernelSelectionMethod::KSM_CROSS_VALIDATION, num_runs, num_folds, alpha);
mmd->select_kernel();
auto learnt_kernel_single = CGaussianKernel::obtain_from_generic(mmd->get_kernel());
auto width = learnt_kernel_single->get_width();
//![select_kernel_single]

//![select_kernel_combined]
mmd->set_kernel_selection_strategy(EKernelSelectionMethod::KSM_MAXIMIZE_MMD, true);
mmd->select_kernel();
auto learnt_kernel_combined = CCombinedKernel::obtain_from_generic(mmd->get_kernel());
auto weights = learnt_kernel_combined->get_subkernel_weights();
//![select_kernel_combined]

//![perform_test_optimized]
auto statistic_optimized = mmd->compute_statistic();
auto p_value_optimized = mmd->compute_p_value(statistic);
//![perform_test_optimized]

// Serialize output for integration testing (automatically generated)
auto __sg_storage = some<CWrappedObjectArray>();
auto __sg_storage_file = some<CSerializableAsciiFile>("quadratic_time_mmd.dat", 'w');
__sg_storage->append_wrapped_real(alpha, "alpha");
__sg_storage->append_wrapped_real(statistic, "statistic");
__sg_storage->append_wrapped_real(threshold, "threshold");
__sg_storage->append_wrapped_real(p_value, "p_value");
__sg_storage->append_wrapped_real_vector(mk_statistic, "mk_statistic");
__sg_storage->append_wrapped_real_vector(mk_p_value, "mk_p_value");
__sg_storage->append_wrapped_int(num_runs, "num_runs");
__sg_storage->append_wrapped_int(num_folds, "num_folds");
__sg_storage->append_wrapped_real(width, "width");
__sg_storage->append_wrapped_real_vector(weights, "weights");
__sg_storage->append_wrapped_real(statistic_optimized, "statistic_optimized");
__sg_storage->append_wrapped_real(p_value_optimized, "p_value_optimized");
__sg_storage->save_serializable(__sg_storage_file);

exit_shogun();
return 0;
}
